/*
 *  CPE 471 Final Project -- Flashlight Sliding Maze
 *  glut/OpenGL/GLSL application 
 *
 *  Created by Brandon Bussjaeger on 6/6/12
 *  bbussjae@calpoly.edu 
 *
 *****************************************************************************/
#ifdef __APPLE__
#include "GLUT/glut.h"
#include <OPENGL/gl.h>
#endif
#ifdef __unix__
#include <GL/glut.h>
#endif

#include "CMeshLoaderSimple.h"
#include <stdlib.h>
#include <stdio.h>
#include "GLSL_helper.h"
#include "glm/glm.hpp"
#include "glm/gtc/matrix_transform.hpp" //perspective, trans etc
#include "glm/gtc/type_ptr.hpp" //value_ptr
#include "RenderingHelper.h"

using namespace std;
using namespace glm;

#define PI 3.141592654

#define MESH_GROUND 0
#define MESH_ROCK 1
#define MESH_WALL 2
#define MESH_TYRA 3

// speed of camera turn
#define ANIMATION_TIMER 10

//flag and ID to toggle on and off the shader
int shade = 1;
int ShadeProg;

//Handles to the shader data
GLint h_aPosition;
GLint h_aNormal;
        
GLint h_uProjMatrix;
GLint h_uViewMatrix;
GLint h_uModelMatrix;
  	  
GLint h_uLightPos;
GLint h_uLightColor;
GLint h_uLightDir;

GLint h_uFlarePos;
GLint h_uFlareBrightness;

GLint h_uBrightness;
GLint h_uGodView;
        
GLint h_uMatAmb;
GLint h_uMatDif;
GLint h_uMatSpec;
GLint h_uMatShine;

GLuint GrndBuffObj, GIndxBuffObj, GroundNormalBO;
int g_CiboLen, g_GiboLen;

GLuint BunnyIBO, BunnyVBO, BunnyNormalBO;
int BunnyTriangleCount;

GLuint TyraIBO, TyraVBO, TyraNormalBO;
int TyraTriangleCount;

RenderingHelper ModelTrans;


// Window paramaters
static float g_width, g_height;

#define START_CAM_X 14.0
#define START_CAM_Y 0.1
#define START_CAM_Z 14.3

// Camera and look variables
float g_camX = START_CAM_X;
float g_camY = START_CAM_Y;
float g_camZ = START_CAM_Z;
float g_camLookX = 0.0;
float g_camLookY = 0.0;
float g_camLookZ = -1.0;
float g_camTheta = 0.0; // looking down, from +X axis, clockwise 
float g_camPhi = 0.0; // from the ground plane. angle looking up is positive 

// Animation variables
float g_camNextTheta = 0.0;
float g_camNextX = START_CAM_X;
float g_camNextZ = START_CAM_Z;
float g_isSliding = 0.0;

float g_flareBrightness = 1.0;
float g_flareX = 15.0;
float g_flareY = -0.7;
float g_flareZ = 7.5;

#define CAM_SLIDE_SPEED .1

// Mostly cheats
float g_noClip = 0;
float g_godView = 0;
float g_brightness = 3.0;

// Constants for ground
static const float g_groundY = -0.6;      // y coordinate of the ground
static const float g_groundSize = 60.0;   // half the ground length


// g_camZ is vertical position
// g_camX is horizontal position
int rockLocations[16][16] = {1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
                             1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1,
                             1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
                             1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1,
                             1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,
                             1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1,
                             1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1,
                             1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, // End by running into this
                             1, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, // or this
                             1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1,
                             1, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1,
                             1, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 1,
                             1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1,                             
                             1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 1, //start at the 0
                             1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
                             1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,}; 


// Forward function declarations
void noClipKeyboard(unsigned char key, int x, int y );
void slidingKeyboard(unsigned char key, int x, int y );
void mouseMove (int x, int y);

/* projection matrix */
void SetProjectionMatrix() {
  //glm::perspective(fovY, aspectRatio, nearZ, farZ);
  glm::mat4 Projection = glm::perspective(80.0f, (float)g_width/g_height, 0.001f, 20.f);	
  safe_glUniformMatrix4fv(h_uProjMatrix, glm::value_ptr(Projection));
}

/* camera controls */
void SetView() {   
  mat4 Camera;
  if (g_godView == 0) {
    safe_glUniform1f(h_uGodView, 0.0);
    g_camZ--;
    Camera = lookAt(vec3(g_camX, g_camY, g_camZ), 
                         vec3(g_camLookX+g_camX, g_camLookY+g_camY, g_camLookZ+g_camZ),
                         vec3( 0.0, 1.0, 0.0));
    g_camZ++;                     
  } else {
    safe_glUniform1f(h_uGodView, 1.0);
    g_camZ--;
    Camera = lookAt(vec3(7.8, 9.6, 9.3), 
                         vec3(7.8, 8.6, 9.27),
                         vec3( 0.0, 1.0, 0.0));
    g_camZ++;
  }
                       
  safe_glUniformMatrix4fv(h_uViewMatrix, glm::value_ptr(Camera));
  glutWarpPointer(g_width/2, g_height/2);
  glutPostRedisplay();

}

void SetModel() {
  safe_glUniformMatrix4fv(h_uModelMatrix, glm::value_ptr(ModelTrans.modelViewMatrix));
}

/* intialize ground data */
static void initGround() {

  // A x-z plane at y = g_groundY of dimension [-g_groundSize, g_groundSize]^2
    float GrndPos[] = { 
    -g_groundSize, g_groundY, -g_groundSize, 
    -g_groundSize, g_groundY,  g_groundSize, 
     g_groundSize, g_groundY,  g_groundSize, 
     g_groundSize, g_groundY, -g_groundSize
    };
    
    float GrndNorm[] = {0, 1, 0, 0, 1, 0};
    unsigned short idx[] = {0, 1, 2, 0, 2, 3};
    
    static GLfloat CubeTex[] = {
      0, 0, // one triangle 
      0, 1,
      1, 1,
      1, 0,
      0, 0, // other triangle 
      0, 1,
      1, 1,
      1, 0
    };
  
    g_GiboLen = 6;
    glGenBuffers(1, &GrndBuffObj);
    glBindBuffer(GL_ARRAY_BUFFER, GrndBuffObj);
    glBufferData(GL_ARRAY_BUFFER, sizeof(GrndPos), GrndPos, GL_STATIC_DRAW);
    
    glGenBuffers(1, &GroundNormalBO);
    glBindBuffer(GL_ARRAY_BUFFER, GroundNormalBO);
    glBufferData(GL_ARRAY_BUFFER, sizeof(GrndNorm), GrndNorm, GL_STATIC_DRAW);

    glGenBuffers(1, &GIndxBuffObj);
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, GIndxBuffObj);
    glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(idx), idx, GL_STATIC_DRAW);
}

// load in models
void InitGeom() {
  initGround();
  CMeshLoader::loadVertexBufferObjectFromMesh("Models/bunny500.m", BunnyTriangleCount, BunnyVBO, BunnyIBO, BunnyNormalBO);
  CMeshLoader::loadVertexBufferObjectFromMesh("Models/tyra_20k.m", TyraTriangleCount, TyraVBO, TyraIBO, TyraNormalBO);
}

/*function to help load the shaders (both vertex and fragment */
int InstallShader(const GLchar *vShaderName, const GLchar *fShaderName) {
        GLuint VS; //handles to shader object
        GLuint FS; //handles to frag shader object
        GLint vCompiled, fCompiled, linked; //status of shader

        VS = glCreateShader(GL_VERTEX_SHADER);
        FS = glCreateShader(GL_FRAGMENT_SHADER);

        //load the source
        glShaderSource(VS, 1, &vShaderName, NULL);
        glShaderSource(FS, 1, &fShaderName, NULL);

        //compile shader and print log
        glCompileShader(VS);
        /* check shader status requires helper functions */
        printOpenGLError();
        glGetShaderiv(VS, GL_COMPILE_STATUS, &vCompiled);
        printShaderInfoLog(VS);

        //compile shader and print log
        glCompileShader(FS);
        /* check shader status requires helper functions */
        printOpenGLError();
        glGetShaderiv(FS, GL_COMPILE_STATUS, &fCompiled);
        printShaderInfoLog(FS);

        if (!vCompiled || !fCompiled) {
                printf("Error compiling either shader %s or %s", vShaderName, fShaderName);
                return 0;
        }

        //create a program object and attach the compiled shader
        ShadeProg = glCreateProgram();
        glAttachShader(ShadeProg, VS);
        glAttachShader(ShadeProg, FS);

        glLinkProgram(ShadeProg);
        /* check shader status requires helper functions */
        printOpenGLError();
        glGetProgramiv(ShadeProg, GL_LINK_STATUS, &linked);
        printProgramInfoLog(ShadeProg);

        glUseProgram(ShadeProg);

        /* get handles to attribute data */
        h_aPosition = safe_glGetAttribLocation(ShadeProg, "aPosition");
        h_aNormal = safe_glGetAttribLocation(ShadeProg, "aNormal");
        
    	  h_uProjMatrix = safe_glGetUniformLocation(ShadeProg, "uProjMatrix");
    	  h_uViewMatrix = safe_glGetUniformLocation(ShadeProg, "uViewMatrix");
    	  h_uModelMatrix = safe_glGetUniformLocation(ShadeProg, "uModelMatrix");
    	  
    	  h_uLightPos = safe_glGetUniformLocation(ShadeProg, "uLightPos");
        h_uLightColor = safe_glGetUniformLocation(ShadeProg, "uLColor");
        h_uLightDir = safe_glGetUniformLocation(ShadeProg, "uLightDir");
        
        h_uFlarePos = safe_glGetUniformLocation(ShadeProg, "uFlarePos");
        h_uFlareBrightness = safe_glGetUniformLocation(ShadeProg, "uFlareBrightness");
        
        h_uBrightness = safe_glGetUniformLocation(ShadeProg, "uBrightness");
        h_uGodView = safe_glGetUniformLocation(ShadeProg, "uGodView");
        
        h_uMatAmb = safe_glGetUniformLocation(ShadeProg, "uMat.aColor");
        h_uMatDif = safe_glGetUniformLocation(ShadeProg, "uMat.dColor");
        h_uMatSpec = safe_glGetUniformLocation(ShadeProg, "uMat.sColor");
        h_uMatShine = safe_glGetUniformLocation(ShadeProg, "uMat.shine");

        printf("sucessfully installed shader %d\n", ShadeProg);
        return 1;
	
}


/* Some OpenGL initialization */
void Initialize ()
{
	// Start Of User Initialization
	glClearColor (.0f, .0f, .0f, 1.0f);		
   ModelTrans.useModelViewMatrix();
   ModelTrans.loadIdentity();						
	// Black Background
 	glClearDepth (1.0f);	// Depth Buffer Setup
 	glDepthFunc (GL_LEQUAL);	// The Type Of Depth Testing
	glEnable (GL_DEPTH_TEST);// Enable Depth Testing
}

void SetMesh(int s) {
   switch (s) {
   case MESH_GROUND:
      glUniform3f(h_uMatAmb, 0.1, 0.1, 0.13);
      glUniform3f(h_uMatDif, 1.0, 1.0, 2.0);
      glUniform3f(h_uMatSpec, 0.7, 0.7, 0.7);
      glUniform1f(h_uMatShine, .0001); 
      break;
   case MESH_ROCK:
      glUniform3f(h_uMatAmb, 0.0698, 0.06117, 0.02078);
      glUniform3f(h_uMatDif, .698, .6117, .2078);
      glUniform3f(h_uMatSpec, 0.35, 0.306, 0.1039	);
      glUniform1f(h_uMatShine, 1.0);
      break;
   case MESH_WALL: 
      glUniform3f(h_uMatAmb, .05, .05, .05);
      glUniform3f(h_uMatDif, .1, .1, .1);
      glUniform3f(h_uMatSpec, 0.1, 0.1, 0.1	);
      glUniform1f(h_uMatShine, 10.0);
      break;
   case MESH_TYRA:
      glUniform3f(h_uMatAmb, .3, .45, .15);
      glUniform3f(h_uMatDif, .4, 1.0, .4);
      glUniform3f(h_uMatSpec, 0.1, 0.1, 0.1	);
      glUniform1f(h_uMatShine, .01);
      break;
   }
}


// draws a bunny objectwith the current model matrix
// does not call SetMesh as bunnies have different meshes for walls or center
void DrawBunny() {

   safe_glEnableVertexAttribArray(h_aPosition);
 	glBindBuffer(GL_ARRAY_BUFFER, BunnyVBO);
	safe_glVertexAttribPointer(h_aPosition, 3, GL_FLOAT, GL_FALSE, 0, 0);
	
	safe_glEnableVertexAttribArray(h_aNormal);
 	glBindBuffer(GL_ARRAY_BUFFER, BunnyNormalBO);
	safe_glVertexAttribPointer(h_aNormal, 3, GL_FLOAT, GL_FALSE, 0, 0);

   glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, BunnyIBO);
   glDrawElements(GL_TRIANGLES, BunnyTriangleCount*3, GL_UNSIGNED_INT, 0);

	//clean up 
	safe_glDisableVertexAttribArray(h_aPosition);
	safe_glDisableVertexAttribArray(h_aNormal);
}

void DrawTyra() {

   SetMesh(MESH_TYRA);

   safe_glEnableVertexAttribArray(h_aPosition);
 	glBindBuffer(GL_ARRAY_BUFFER, TyraVBO);
	safe_glVertexAttribPointer(h_aPosition, 3, GL_FLOAT, GL_FALSE, 0, 0);
	
	safe_glEnableVertexAttribArray(h_aNormal);
 	glBindBuffer(GL_ARRAY_BUFFER, TyraNormalBO);
	safe_glVertexAttribPointer(h_aNormal, 3, GL_FLOAT, GL_FALSE, 0, 0);

   glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, TyraIBO);
   glDrawElements(GL_TRIANGLES, TyraTriangleCount*3, GL_UNSIGNED_INT, 0);

	//clean up 
	safe_glDisableVertexAttribArray(h_aPosition);
	safe_glDisableVertexAttribArray(h_aNormal);
}


void DrawGround() {

   SetMesh(MESH_GROUND);

   safe_glEnableVertexAttribArray(h_aPosition);
   glBindBuffer(GL_ARRAY_BUFFER, GrndBuffObj);
   safe_glVertexAttribPointer(h_aPosition, 3, GL_FLOAT, GL_FALSE, 0, 0);

   safe_glEnableVertexAttribArray(h_aNormal);
 	glBindBuffer(GL_ARRAY_BUFFER, GroundNormalBO);
	safe_glVertexAttribPointer(h_aNormal, 3, GL_FLOAT, GL_FALSE, 0, 0);

  	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, GIndxBuffObj);
  	glDrawElements(GL_TRIANGLES, g_GiboLen, GL_UNSIGNED_SHORT, 0);
  	
  	// Disable the attributes used by our shader
  	safe_glDisableVertexAttribArray(h_aPosition);
   safe_glDisableVertexAttribArray(h_aNormal);
}

/* Main display function */
void Draw (void) {

   glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);					
   glUseProgram(ShadeProg);

   //Set up the projection and view matricies
   SetProjectionMatrix();
   SetView();
   
   // Set up the light position, color, and direction
   glUniform3f(h_uLightPos, g_camX, g_camY, g_camZ-1.0);
   glUniform3f(h_uLightDir, g_camLookX, g_camLookY, g_camLookZ);
   glUniform3f(h_uLightColor, 1.0, 1.0, 1.0);
   glUniform1f(h_uBrightness, g_brightness);
   
   // Set up the green flare at the finish
   glUniform3f(h_uFlarePos, g_flareX, g_flareY, g_flareZ);
   glUniform1f(h_uFlareBrightness, g_flareBrightness);

   // Draw the Ground    
   SetModel();
   DrawGround();
   
   // Draw a dinosaur where the player is if in godview
   if (g_godView != 0) {
      ModelTrans.pushMatrix();
         ModelTrans.translate(vec3(g_camX, g_camY, g_camZ-1.0));
         SetModel();
         DrawTyra();
      ModelTrans.popMatrix();
   }
      
   // Draw the game map - made of bunnies
   ModelTrans.pushMatrix();
      for (int i = 0; i < 16; i++) {
         for (int j = 0; j < 16; j++) { 
            if (rockLocations[i][j] == 1) {
               ModelTrans.pushMatrix();
                  if (i == 0 || j == 0 || i >= 13 || j == 15) {
                     SetMesh(MESH_WALL);
                  } else {
                     SetMesh(MESH_ROCK);
                  }
                  ModelTrans.translate(vec3(j, 0, i));
                  SetModel();
                  DrawBunny();
               ModelTrans.popMatrix();
            }
         }
      }
   ModelTrans.popMatrix();
   
	//Disable the shader
	glUseProgram(0);	
	glutSwapBuffers();
}

/* Reshape */
void ReshapeGL (int width, int height) {
	g_width = (float)width;
	g_height = (float)height;
	glViewport (0, 0, (GLsizei)(width), (GLsizei)(height));				
}

// randomly vary the flare's brightness to get a flicker effect
void flickerFlare (int trash) {
   g_flareBrightness += (rand() % 2 == 1) ? 0.1 : -0.1;
   if (g_flareBrightness < 0.0) {
     g_flareBrightness = .1;
   }
   if (g_flareBrightness > 3.0) {
     g_flareBrightness = 2.5;
   }
	glutTimerFunc(ANIMATION_TIMER, flickerFlare, 0);
}

// moves the camera view slowly do the desired location.
// Used when the player hits the "w" key to move, but is
// not facing an exact direction.
void animateLook (int trash) {
  
  // figure out which way to turn
  if (g_camTheta > g_camNextTheta) {
     if (g_camNextTheta == 0.0 && g_camTheta > 7.0*PI/4) {
        // special case where we increment to get to 0.0
        g_camTheta += .02;
     } else {
        // no special case, decrememnt camTheta to go to nextTheta
        g_camTheta -= .02;
     }
  } else {
     // else camTheta < nextTheta, so increment
     g_camTheta += .02;
  }

  // if we arent close enough, keep animating
  if (abs(g_camTheta - g_camNextTheta) > .021)
     glutTimerFunc(ANIMATION_TIMER, animateLook, 0);
  else
     g_camTheta = g_camNextTheta;
     
  // call mouse move with no movement to update our camLook<X,Y,Z>
  mouseMove(g_width/2.0, g_height/2.0);
  glutPostRedisplay();
} 

// Moves the player forward until a wall is hit
// gives a sliding motion to simulate ice.
// The player can not command more moves while sliding
void animateSlide (int trash) {

  if (g_camNextX > g_camX) {
     g_camX += CAM_SLIDE_SPEED;
  }
  if (g_camNextX < g_camX) {
     g_camX -= CAM_SLIDE_SPEED;
  }
  if (g_camNextZ > g_camZ) {
     g_camZ += CAM_SLIDE_SPEED;
  }
  if (g_camNextZ < g_camZ) {
     g_camZ -= CAM_SLIDE_SPEED;
  }
  
  if (abs(g_camX-g_camNextX) > CAM_SLIDE_SPEED+.001 || abs(g_camZ-g_camNextZ) > CAM_SLIDE_SPEED+.001) {
     glutTimerFunc(ANIMATION_TIMER, animateSlide, 0);
  } else {
     g_isSliding = 0.0;
     g_camX = g_camNextX;
     g_camZ = g_camNextZ;
  }
}

// main keyboard to call the correct keyboard function depending if noclip is on
void keyboard(unsigned char key, int x, int y) {
   if (g_noClip == 0) 
      slidingKeyboard(key, x, y);
   else 
      noClipKeyboard(key, x, y);
}

// the user keyboard for normal sliding gameplay
void slidingKeyboard(unsigned char key, int x, int y ){
  switch( key ) {
    /* WASD keyes effect view/camera transform */
    case 'w': {
      if (g_isSliding == 1.0) {
         break;
      }
      /* first check if we are facing close enough to an exact direction to move 
         ~~remember~~ g_camTheta is clockwise on the +X axis
         ~~remember~~ in terms of rockLocations[][], you are at [g_camZ-1][g_camX]
      */
      
      // Start camNextX,Z at our current X,Z. Then set them to our final location
      // and call animateSlide(). Animate incememntally sets camX,Z to camNextX,Z
      g_camNextX = g_camX;
      g_camNextZ = g_camZ;
      
      if (g_camTheta > -.1 && g_camTheta < 0.1) { 
        //move camera in the +X direction if no obstacle
        while (rockLocations[(int)g_camNextZ-1][(int)g_camNextX+1] != 1) {
           g_camNextX++;
        }
        g_isSliding = 1.0;
        glutTimerFunc(ANIMATION_TIMER, animateSlide, 0); 
        
      } else if (g_camTheta > 1.5 && g_camTheta < 1.6) {
        //move camera in the +Z direction
        while (rockLocations[(int)g_camNextZ-1+1][(int)g_camNextX] != 1) {
           g_camNextZ++;
        }
        g_isSliding = 1.0;
        glutTimerFunc(ANIMATION_TIMER, animateSlide, 0); 
        
      } else if (g_camTheta > 3.1 && g_camTheta < 3.3) {
        //move camera in the -X direction
        while (rockLocations[(int)g_camNextZ-1][(int)g_camNextX-1] != 1) {
           g_camNextX--;
        }
        g_isSliding = 1.0;
        glutTimerFunc(ANIMATION_TIMER, animateSlide, 0); 
        
      } else if (g_camTheta > 4.6 && g_camTheta < 4.8) {
        //move camera in the -Z direction
        while (rockLocations[(int)g_camNextZ-2][(int)g_camNextX] != 1) {
           g_camNextZ--;
        }
        g_isSliding = 1.0;
        glutTimerFunc(ANIMATION_TIMER, animateSlide, 0);    
      } else
    
      /* otherwise, face the closest exact direction to our current view */
    
      if (g_camTheta > 7.0*PI/4.0 || g_camTheta < PI/4.0) {
         g_camNextTheta = 0.0;
         glutTimerFunc (ANIMATION_TIMER, animateLook, 0);
      } else if (g_camTheta > PI/4.0 && g_camTheta < 3.0*PI/4.0) {
         g_camNextTheta = PI/2.0;
         glutTimerFunc (ANIMATION_TIMER, animateLook, 0);
      } else if (g_camTheta > 3.0*PI/4.0 && g_camTheta < 5.0*PI/4.0) {
         g_camNextTheta = PI;
         glutTimerFunc (ANIMATION_TIMER, animateLook, 0);
      } else if (g_camTheta > 5.0*PI/4.0 && g_camTheta < 7.0*PI/4.0) {
         g_camNextTheta = 3.0*PI/2.0;
         glutTimerFunc (ANIMATION_TIMER, animateLook, 0);
      } else {
         printf("failure %f\n", g_camTheta);
      }
      mouseMove(g_width/2.0, g_height/2.0);
      break;
    }
    case 'n': {
      g_noClip = 1;
      break;
    }
    case 'k': {
      g_brightness--;
      break;
    }
    case 'l': {
      g_brightness++;
      break;
    }
    case 'g': {
      g_godView = (g_godView == 1) ? 0 : 1;
      break;
    }
    case 'q': case 'Q' :
      exit( EXIT_SUCCESS );
      break;
  }
  glutPostRedisplay();
}

// the "noclip" keyboard for a free-flying camera. 
// exiting noclip keyboard mode restarts the game.
void noClipKeyboard(unsigned char key, int x, int y ){
  switch( key ) {
    /* WASD keyes effect view/camera transform */
    case 'w': {
      vec3 la = vec3(g_camLookX+g_camX, g_camLookY+g_camY, g_camLookZ+g_camZ);
      vec3 eye = vec3(g_camX, g_camY, g_camZ);
      vec3 w = la - eye;
      w = normalize(w);
      eye = eye + 0.2f*w;
      g_camX = eye.x;
      g_camY = eye.y;
      g_camZ = eye.z;
      break;
    }
    case 's': {
      vec3 la = vec3(g_camLookX+g_camX, g_camLookY+g_camY, g_camLookZ+g_camZ);
      vec3 eye = vec3(g_camX, g_camY, g_camZ);
      vec3 w = la - eye;
      w = normalize(w);
      eye = eye + -0.2f*w;
      g_camX = eye.x;
      g_camY = eye.y;
      g_camZ = eye.z;
      break;
    }	
    case 'a': {
      vec3 la = vec3(g_camLookX+g_camX, g_camLookY+g_camY, g_camLookZ+g_camZ);
      vec3 eye = vec3(g_camX, g_camY, g_camZ);
      vec3 w = la - eye;
      w = normalize(w);
      vec3 u = cross(vec3(0, 1, 0), w);
      u = normalize(u);      
      eye = eye + 0.2f*u;
      g_camX = eye.x;
      g_camY = eye.y;
      g_camZ = eye.z;
      break;
    }
    case 'd': {
      vec3 la = vec3(g_camLookX+g_camX, g_camLookY+g_camY, g_camLookZ+g_camZ);
      vec3 eye = vec3(g_camX, g_camY, g_camZ);
      vec3 w = la - eye;
      w = normalize(w);
      vec3 u = cross(vec3(0, 1, 0), w);
      u = normalize(u);      
      eye = eye + -0.2f*u;
      g_camX = eye.x;
      g_camY = eye.y;
      g_camZ = eye.z;
      break;
    }
    case 'n': {
       g_noClip = 0;
       g_camX = START_CAM_X;
       g_camY = START_CAM_Y;
       g_camZ = START_CAM_Z;
       break;
    }
    case 'g': {
       g_godView = (g_godView == 1) ? 0 : 1;
       break;
    }
    case 'q': case 'Q' :
      exit( EXIT_SUCCESS );
      break;
  }
  glutPostRedisplay();
}

float p2wx(int in_x) {
  return ((2/(float)(g_width-1))*in_x)-1;
}

float p2wy(int in_y) {
  //flip glut y
  in_y = g_height - in_y;
  return ((2/(float)(g_height-1))*in_y)-1;
}

void mouse(int button, int state, int x, int y) { }

// the mouse is always recentered, and the cursor is not shown.
// find the move from center every time, ignore "last position" math
void mouseMove (int x, int y) {
   float currX = p2wx(x);
   float currY = p2wy(y);
   g_camTheta += .5 * currX;
   g_camPhi += .5 * currY;
   
   while (g_camTheta > 2.0*PI) { g_camTheta -= 2.0*PI; }
   while (g_camTheta < 0.0)    { g_camTheta += 2.0*PI; }

   if (g_camPhi > 85.0*PI/180.0)  { g_camPhi = 85.0*PI/180.0; }
   if (g_camPhi < -85.0*PI/180.0) { g_camPhi = -85.0*PI/180.0; }
   
   g_camLookX = cos(g_camPhi)*cos(g_camTheta);
   g_camLookY = sin(g_camPhi); 
   g_camLookZ = cos(g_camPhi)*cos(PI/2.0-g_camTheta);
}

int main( int argc, char *argv[] ) {
    
   	glutInit( &argc, argv );
   	glutInitWindowPosition( 20, 20 );
   	glutInitWindowSize( 800, 600 );
   	glutInitDisplayMode( GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH );
   	glutCreateWindow("Maze of evil bunnys!");
   	glutFullScreen();
   	glutSetCursor(GLUT_CURSOR_NONE); 
   	glutReshapeFunc( ReshapeGL );
   	glutDisplayFunc( Draw );
	   glutKeyboardFunc( keyboard );
	   glutMouseFunc( mouse );
	   glutMotionFunc( mouseMove );
	   glutPassiveMotionFunc( mouseMove );
   	Initialize();
   	glutTimerFunc(ANIMATION_TIMER, flickerFlare, 0);

	//test the openGL version
	getGLversion();
	//install the shader
	if (!InstallShader(textFileRead((char *)"final_vert.glsl"), textFileRead((char *)"final_frag.glsl"))) {
		printf("Error installing shader!\n");
		return 0;
	}
		
	InitGeom();
  	glutMainLoop();
   	return 0;
}
